/* AJet.h
 *
 * Written by: Peter Milley
 *
 * An Affine1Jet represents a set of functions f with domain the
 * standard cube in R^3 (i.e. the cube {|x|<=1, |y|<=1, |z|<=1}), such
 * that |f(v)-f0-(df1,df2,df3).v| < error for all v and f.
 */

#ifndef _AJet_h_
#define _AJet_h_

#include <math.h> // for fabs, log

class Affine1Jet
{
public:
  // Constructors and destructors
  Affine1Jet( double f0, double df1=0, double df2=0, double df3=0,
	      double e=0 );
  Affine1Jet();
  Affine1Jet( const Affine1Jet & );
  Affine1Jet & operator= ( const Affine1Jet & );

  ~Affine1Jet() { }; // empty

  // Useful constants
  static const double EPS;
  static const double EPS_by_2;

  // Exception classes.
  class DivisionByZeroError {
  public:
    DivisionByZeroError() {} // empty;
  };
  class ZeroToTheZeroError {
  public:
    ZeroToTheZeroError() {} // empty;
  };
  class LogZeroError {
  public:
    LogZeroError() {} // empty;
  };
  class LogIsComplexError {
  public:
    LogIsComplexError() {} // empty;
  };

  // Access functions
  inline double center() const;
  inline double size() const;
  inline double slope( int ) const;
  inline double error() const;

  // Some "arithmetic functions" defined specially.
  Affine1Jet & squeezeLeft( int index );
  Affine1Jet & squeezeRight( int index );
  Affine1Jet & flatten();

  // Arithmetic assignment operators.
  Affine1Jet & operator+= ( const Affine1Jet & );
  Affine1Jet & operator+= ( double );
  Affine1Jet & operator-= ( const Affine1Jet & );
  Affine1Jet & operator-= ( double );
  Affine1Jet & operator*= ( const Affine1Jet & );
  Affine1Jet & operator*= ( double );
  Affine1Jet & operator/= ( const Affine1Jet & );
  Affine1Jet & operator/= ( double );

  // A test function. Returns true if every point in the range of the
  // jet is greater than the given argument.
  // Not defined as an operator; the semantics aren't quite right.
  bool trulyGreaterThan( double d ) const;
  bool trulyGreaterThan( const Affine1Jet & w ) const;

  // Arithmetic operators.
  friend Affine1Jet operator+ ( const Affine1Jet &, const Affine1Jet & );
  friend Affine1Jet operator+ ( const Affine1Jet &, double );
  friend Affine1Jet operator+ ( double, const Affine1Jet & );

  friend Affine1Jet operator- ( const Affine1Jet &, const Affine1Jet & );
  friend Affine1Jet operator- ( const Affine1Jet &, double );
  friend Affine1Jet operator- ( double, const Affine1Jet & );

  friend Affine1Jet operator- ( const Affine1Jet & );

  friend Affine1Jet operator* ( const Affine1Jet &, const Affine1Jet & );
  friend Affine1Jet operator* ( const Affine1Jet &, double );
  friend Affine1Jet operator* ( double, const Affine1Jet & );

  friend Affine1Jet operator/ ( const Affine1Jet &, const Affine1Jet & );
  friend Affine1Jet operator/ ( const Affine1Jet &, double );
  friend Affine1Jet operator/ ( double, const Affine1Jet & );

  // Test for equality. Nothing approximate here; this is exact equality.
  friend bool operator==( const Affine1Jet &, const Affine1Jet & );

  // Other useful mathematical functions.
  friend Affine1Jet Power ( const Affine1Jet &, int );
  friend Affine1Jet Log ( Affine1Jet );
  friend Affine1Jet Max0 ( const Affine1Jet & );

  friend Affine1Jet ArcTan ( Affine1Jet );

private:
  bool mayContainZero() const;

  double _f0;
  double _df1, _df2, _df3;
  double _error;
  double _size;
};

#include "AJet.inline"

#endif // _AJet_h_
