/* AJet.C
 *
 * Written by: Peter Milley
 * Inspired heavily by existing code from Nathaniel Thurston
 */

#include "AJet.H"
#include <iostream>

// EPS refers to the smallest positive real number such that 1+EPS has
// a floating-point representation distinct from 1.
#include <float.h>
const double Affine1Jet::EPS = DBL_EPSILON;
const double Affine1Jet::EPS_by_2 = EPS / 2;

Affine1Jet::Affine1Jet( double f0, double df1, double df2,
			double df3, double e )
{
  _f0 = f0;
  _df1 = df1;
  _df2 = df2;
  _df3 = df3;
  _error = e;
}

Affine1Jet::Affine1Jet( )
{
  _f0 = 0.0;
  _df1 = 0.0;
  _df2 = 0.0;
  _df3 = 0.0;
  _error = 0.0;
}

Affine1Jet::Affine1Jet( const Affine1Jet & j )
{
  _f0 = j._f0;
  _df1 = j._df1;
  _df2 = j._df2;
  _df3 = j._df3;
  _error = j._error;
}

Affine1Jet & Affine1Jet::operator=( const Affine1Jet & j )
{
  _f0 = j._f0;
  _df1 = j._df1;
  _df2 = j._df2;
  _df3 = j._df3;
  _error = j._error;

  return (*this);
}

// Arithmetic assignment operators are all defined in terms of the
// corresponding arithmetic operator.

Affine1Jet & Affine1Jet::operator+= ( const Affine1Jet & w )
{
  Affine1Jet result = (*this) + w;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator+= ( double d )
{
  Affine1Jet result = (*this) + d;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator-= ( const Affine1Jet & w )
{
  Affine1Jet result = (*this) - w;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator-= ( double d )
{
  Affine1Jet result = (*this) - d;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator*= ( const Affine1Jet & w )
{
  Affine1Jet result = (*this) * w;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator*= ( double d )
{
  Affine1Jet result = (*this) * d;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator/= ( const Affine1Jet & w )
{
  Affine1Jet result = (*this) / w;
  (*this) = result;
  return (*this);
}

Affine1Jet & Affine1Jet::operator/= ( double d )
{
  Affine1Jet result = (*this) / d;
  (*this) = result;
  return (*this);
}

/* trulyGreaterThan returns "true" only if the range of the 1-jet
 * definitely lies strictly above the given argument, even given the
 * floating-point error which occurs during the computation of that
 * range. It assumes the argument (if a double) is given exactly.
 */
bool Affine1Jet::trulyGreaterThan( double d ) const
{
  double boundary_est = d + ((fabs(_df1) + fabs(_df2)) +
			     (fabs(_df3) + _error));
  // Account for the floating point error in the above calculation.
  if( boundary_est >= 0 ){
    boundary_est *= 1.0 + 3 * EPS;
  } else {
    boundary_est *= 1.0 - 3 * EPS;
  }

  return boundary_est < _f0;
}

bool Affine1Jet::trulyGreaterThan( const Affine1Jet & w ) const
{
  double boundary_est = w._f0 + (((fabs(w._df1) + fabs(w._df2)) +
				  (fabs(w._df3) + w._error)) +
				 ((fabs(_df1) + fabs(_df2)) +
				  (fabs(_df3) + _error)));
  // Account for the floating point error in the above calculation.
  if( boundary_est >= 0 ){
    boundary_est *= 1.0 + 3 * EPS;
  } else {
    boundary_est *= 1.0 - 3 * EPS;
  }

  return boundary_est < _f0;
}
    

/* mayContainZero returns "false" only if the range of the 1-jet
 * definitely does NOT contain zero, even given the floating-point
 * error which occurs during the computation of that range.
 */
bool Affine1Jet::mayContainZero() const
{
  // Compute the 'diameter' of the 1-jet.
  double diameter_est = (fabs(_df1) + fabs(_df2)) + (fabs(_df3) + _error);
  // Account for the floating-point error in the above calculation.
  diameter_est *= 1.0 + 2 * EPS;

  if( fabs(_f0) <= diameter_est ){
    return true;
  }
  return false;
}

// squeezeLeft() and squeezeRight() return the "left" or "right" half
// of the Affine1Jet, with respect to the dimension given by "index".

Affine1Jet & Affine1Jet::squeezeLeft( int index )
{
  if( index < 0 || index >= 3 ){
    return (*this);
  }

  // The "Taylor" error remains unchanged.
  double error = _error;

  // Add the floating-point error.
  switch( index ){
  case 0:
    error += EPS_by_2 * fabs(_f0 - 0.5 * _df1);
    break;
  case 1:
    error += EPS_by_2 * fabs(_f0 - 0.5 * _df2);
    break;
  case 2:
    error += EPS_by_2 * fabs(_f0 - 0.5 * _df3);
    break;
  }

  // Account for error while calculating the error
  error *= 1.0 + EPS;

  // Now update the jet. The linear term corresponding to index is
  // divided by two and then subtracted from the constant term.
  switch( index ){
  case 0:
    _df1 *= 0.5;
    _f0 -= _df1;
    break;
  case 1:
    _df2 *= 0.5;
    _f0 -= _df2;
    break;
  case 2:
    _df3 *= 0.5;
    _f0 -= _df3;
    break;
  }

  _error = error;
  
  return (*this);
}

Affine1Jet & Affine1Jet::squeezeRight( int index )
{
  if( index < 0 || index >= 3 ){
    return (*this);
  }

  // The "Taylor" error remains unchanged.
  double error = _error;

  // Add the floating-point error.
  switch( index ){
  case 0:
    error += EPS_by_2 * fabs(_f0 + 0.5 * _df1);
    break;
  case 1:
    error += EPS_by_2 * fabs(_f0 + 0.5 * _df2);
    break;
  case 2:
    error += EPS_by_2 * fabs(_f0 + 0.5 * _df3);
    break;
  }

  // Account for error while calculating the error
  error *= 1.0 + EPS;

  // Now update the jet. The linear term corresponding to index is
  // divided by two and then added to the constant term.
  switch( index ){
  case 0:
    _df1 *= 0.5;
    _f0 += _df1;
    break;
  case 1:
    _df2 *= 0.5;
    _f0 += _df2;
    break;
  case 2:
    _df3 *= 0.5;
    _f0 += _df3;
    break;
  }

  _error = error;

  return (*this);
}

// "flatten" turns the Affine1Jet from a small neighbourhood of a 3-D
// interval into a larger neighbourhood of the point at the centre of
// that interval. This represents an enormous loss of accuracy but is
// occasionally mathematically necessary.

Affine1Jet & Affine1Jet::flatten()
{
  double error = (fabs(_df1) + fabs(_df2)) + (fabs(_df3) + _error);

  // Account for floating-point error in this calculation
  error *= 1.0 + 2 * EPS;

  _df1 = _df2 = _df3 = 0.0;
  _error = error;

  return (*this);
}

// Arithmetic operators. In each case, the hardest part is calculating
// the new error term. For each function, the error term consists of:
//  1) The error inherited from the operands,
//  2) The error introducted by taking a Taylor approximation to the
//     result of the operation,
//  (#1 and #2 are grouped together and called the "Taylor error" in
// the calculations below)
//  3) The floating-point error, typically proportional to the sum of
//     the absolute values of all the intermediate results from step
//     1, step 2, and from computing the actual result of the
//     operation,
//  4) The floating-point error involved in calculating the
//     floating-point error. This last error is accounted for by
//     multiplying the error term to that point by (1+n*EPS) where n
//     is roughly proportional to the logarithm of the number of
//     operations used up to that point.
//
// For details of how to calculate the error term in steps 3 and 4,
// see "Homotopy hyperbolic 3-manifolds are hyperbolic" by Gabai,
// Meyerhoff, and Thurston, Annals of Mathematics 157 (2003)
// pp335--431.
//
// One consequence of the method used to calculate the error term in
// step 4 is that the precise order of operations is VERY
// IMPORTANT. What may seem like excessive use of parentheses is
// actually critical to ensuring that the error term calculated in
// step 4 is correct.

Affine1Jet operator+( const Affine1Jet & v, const Affine1Jet & w )
{
  // First, compute the "Taylor error".
  double error = (v._error + w._error);
  // Then, add the floating-point error.
  error += Affine1Jet::EPS_by_2 *
    ((fabs(v._f0 + w._f0) + fabs(v._df1 + w._df1)) +
     (fabs(v._df2 + w._df2) + fabs(v._df3 + w._df3)));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 3 * Affine1Jet::EPS;

  return Affine1Jet(v._f0 + w._f0,
		    v._df1 + w._df1,
		    v._df2 + w._df2,
		    v._df3 + w._df3,
		    error
		    );
}

Affine1Jet operator+( const Affine1Jet & v, double d )
{
  // First, compute the "Taylor error".
  double error = v._error;
  // Then, add the floating-point error.
  error += Affine1Jet::EPS_by_2 * fabs(v._f0 + d);
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + Affine1Jet::EPS;

  return Affine1Jet(v._f0 + d,
		    v._df1, v._df2, v._df3,
		    error
		    );
}

Affine1Jet operator+( double d, const Affine1Jet & v )
{
  return v+d;
}

// unary minus
Affine1Jet operator-( const Affine1Jet & v )
{
  // Since unary negation is an exact operation,
  // we don't have to worry about changes to _error.
  return Affine1Jet(-v._f0, -v._df1, -v._df2, -v._df3, v._error);
}

Affine1Jet operator-( const Affine1Jet & v, const Affine1Jet & w )
{
  return v + (-w);
}

Affine1Jet operator-( const Affine1Jet & v, double d )
{
  return v + (-d);
}

Affine1Jet operator-( double d, const Affine1Jet & v )
{
  return (-v) + d;
}

Affine1Jet operator*( const Affine1Jet & v, const Affine1Jet & w )
{
  // First, compute the "Taylor error".
  double error = (fabs(v._f0) * w._error + fabs(w._f0) * v._error) +
    ((fabs(v._df1) + fabs(v._df2)) + (fabs(v._df3) + v._error)) *
    ((fabs(w._df1) + fabs(w._df2)) + (fabs(w._df3) + w._error));
  // Then, add the floating point error.
  error += Affine1Jet::EPS_by_2 * ((fabs(v._f0 * w._f0) + fabs(v._f0 * w._df1)) +
		       (((fabs(v._df1 * w._f0) +
			  fabs(v._f0 * w._df1 + v._df1 * w._f0)) +
			 (fabs(v._f0 * w._df2) + fabs(v._df2 * w._f0))) +
			((fabs(v._f0 * w._df2 + v._df2 * w._f0) +
			  fabs(v._f0 * w._df3)) +
			 (fabs(v._df3 * w._f0) +
			  fabs(v._f0 * w._df3 + v._df3 * w._f0)))));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 4 * Affine1Jet::EPS;

  return Affine1Jet(v._f0 * w._f0,
		    v._f0 * w._df1 + v._df1 * w._f0,
		    v._f0 * w._df2 + v._df2 * w._f0,
		    v._f0 * w._df3 + v._df3 * w._f0,
		    error
		    );
}

Affine1Jet operator*( const Affine1Jet & v, double d )
{
  // First, compute the "Taylor error".
  double error = fabs(v._error * d);
  // Then, add the floating-point error.
  error += Affine1Jet::EPS_by_2 * ((fabs(v._f0 * d) + fabs(v._df1 * d)) +
		       (fabs(v._df2 * d) + fabs(v._df3 * d)));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 3 * Affine1Jet::EPS;

  return Affine1Jet(v._f0 * d,
		    v._df1 * d,
		    v._df2 * d,
		    v._df3 * d,
		    error
		    );
}

Affine1Jet operator*( double d, const Affine1Jet & v )
{
  return v*d;
}

Affine1Jet operator/( const Affine1Jet & v, const Affine1Jet & w )
{
  // Need to check for division by zero.
  if( w.mayContainZero() ){
    throw Affine1Jet::DivisionByZeroError();
  }

  // First, compute the "Taylor error".
  double error =
    (fabs(v._f0) + ((fabs(v._df1) + fabs(v._df2)) + (fabs(v._df3) + v._error))) /
    (fabs(w._f0) - ((fabs(w._df1) + fabs(w._df2)) + (fabs(w._df3) + w._error))) -
    ((((fabs(v._f0) + fabs(v._df1)) + (fabs(v._df2) + fabs(v._df3))) /
      fabs(w._f0)) +
     (fabs(w._df1) + (fabs(w._df2) + fabs(w._df3))) * (fabs(v._f0) /
     (fabs(w._f0) * fabs(w._f0))));
  // Next, add the floating-point error.
  error += Affine1Jet::EPS_by_2 * ((fabs(v._f0 / w._f0) +
		       ((fabs(v._df1 / w._f0) +
		       2.0 * (1.0 + 2 * Affine1Jet::EPS) *
		       (fabs(v._f0 * w._df1) / fabs(w._f0 * w._f0))) +
		       (fabs((v._f0 * w._df1) / (w._f0 * w._f0)) +
		       fabs(v._df1/w._f0 - (v._f0*w._df1)/(w._f0*w._f0))))) +
		       (((fabs(v._df2 / w._f0) +
		       2.0 * (1.0 + 2 * Affine1Jet::EPS) *
		       (fabs(v._f0 * w._df2) / fabs(w._f0 * w._f0))) +
		       (fabs((v._f0 * w._df2) / (w._f0 * w._f0)) +
		       fabs(v._df2/w._f0 - (v._f0*w._df2)/(w._f0*w._f0)))) +
		       ((fabs(v._df3 / w._f0) +
		       2.0 * (1.0 + 2 * Affine1Jet::EPS) *
		       (fabs(v._f0 * w._df3) / fabs(w._f0 * w._f0))) +
		       (fabs((v._f0 * w._df3) / (w._f0 * w._f0)) +
		       fabs(v._df3/w._f0 - (v._f0*w._df3)/(w._f0*w._f0))))));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 4 * Affine1Jet::EPS;

  return Affine1Jet(v._f0 / w._f0,
		    -v._f0 * w._df1 / (w._f0 * w._f0) + v._df1 / w._f0,
		    -v._f0 * w._df2 / (w._f0 * w._f0) + v._df2 / w._f0,
		    -v._f0 * w._df3 / (w._f0 * w._f0) + v._df3 / w._f0,
		    error
		    );
}

Affine1Jet operator/( const Affine1Jet & v, double d )
{
  // Check for division by zero.
  if( d == 0.0 ){
    throw Affine1Jet::DivisionByZeroError();
  }
  // Compute the "Taylor error".
  double error = v._error / fabs(d);
  // Next, add the floating-point error.
  error += Affine1Jet::EPS_by_2 * ((fabs(v._f0 / d) + fabs(v._df1 / d)) +
		       (fabs(v._df2 / d) + fabs(v._df3 / d)));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 3 * Affine1Jet::EPS;

  return Affine1Jet(v._f0 / d,
		    v._df1 / d,
		    v._df2 / d,
		    v._df3 / d,
		    error
		    );
}

Affine1Jet operator/( double d, const Affine1Jet & w )
{
  // Check for division by zero.
  if( w.mayContainZero() ){
    throw Affine1Jet::DivisionByZeroError();
  }

  // First, compute the "Taylor error".
  double error = fabs(d) / (fabs(w._f0) - ((fabs(w._df1) + fabs(w._df2)) +
			   (fabs(w._df2) + w._error))) -
    (fabs(d) / fabs(w._f0) +
    (fabs(w._df1) + (fabs(w._df2) + fabs(w._df3))) * (fabs(d) / 
    (fabs(w._f0) * fabs(w._f0))));
  // Then, add the floating point error.
  error += Affine1Jet::EPS_by_2 * ((fabs(d/ w._f0) +
		       (fabs((d * w._df1) / (w._f0 * w._f0)) +
		       2 * (1 + 2 * Affine1Jet::EPS) *
			(fabs(d * w._df1) / fabs(w._f0 * w._f0)))) +
		       ((fabs((d * w._df2) / (w._f0 * w._f0)) +
		       2 * (1 + 2 * Affine1Jet::EPS) *
			 (fabs(d * w._df2) / fabs(w._f0 * w._f0))) +
		       (fabs((d * w._df3) / (w._f0 * w._f0)) +
		       2 * (1 + 2 * Affine1Jet::EPS) *
			(fabs(d * w._df3) / fabs(w._f0 * w._f0)))));
  // Finally, account for the error from adding up the errors.
  error *= 1.0 + 4 * Affine1Jet::EPS;

  return Affine1Jet(d / w._f0,
		    -d * w._df1 / (w._f0 * w._f0),
		    -d * w._df2 / (w._f0 * w._f0),
		    -d * w._df3 / (w._f0 * w._f0),
		    error
		    );
}

bool operator==( const Affine1Jet & v, const Affine1Jet & w )
{
  // Just test for strict equality of all terms.
  return (v._f0 == w._f0)
    && (v._df1 == w._df1)
    && (v._df2 == w._df2)
    && (v._df3 == w._df3)
    && (v._error == w._error);
}

// The remaining functions are not rational polynomials in their
// arguments, and hence computing them rigorously is a bit more involved.

Affine1Jet Power( const Affine1Jet & v, int p )
{
  if( p == 0 ){
    // Check that we're not computing 0^0.
    if( v.mayContainZero() ){
      throw Affine1Jet::ZeroToTheZeroError();
    }
    return Affine1Jet(1.0);
  }

  if( p < 0 ){
    return Power(1.0 / v, -p);
  }
  if( p == 1 ){
    return v;
  }

  // It would be nice to compute this in a more direct manner
  // that doesn't cause the error term to balloon, but for now we'll
  // just try to multiply the power out in an efficient way. (By
  // 'efficient' here I mean 'using the fewest multiplications'.)

  int half_p = p / 2;
  Affine1Jet temp = Power(v, half_p);

  if( p % 2 == 1 ){
    return temp * (temp * v);
  } else {
    return temp * temp;
  }
}

Affine1Jet Log( Affine1Jet v )
{
  // Logarithm is obviously not a rational polynomial. However, its
  // derivatives are all rational polynomials. So to compute a
  // "rigorous" logarithm, we use a Taylor approximation. The
  // difference between log and the approximation (which is bounded
  // above thanks to Taylor's theorem) is included in the resulting
  // error term.
  
  // Can't take the log of zero.
  if( v.mayContainZero() ){
    throw Affine1Jet::LogZeroError();
  }

  // This is tricky. First, make sure we aren't taking the log
  // of a negative number. I.e., check to see if:
  //  v._f0 - sum(|v._dfi|) - v._error <= 0

  double diameter_est = (fabs(v._df1) + fabs(v._df2)) + (fabs(v._df3) + v._error);
  if( v._f0 <= (1.0 + 2 * Affine1Jet::EPS) * diameter_est ){
    throw Affine1Jet::LogIsComplexError();
  }

  // Since we're only using a crude polynomial approximation to log,
  // it helps to perform some elaborate chicanery to try and ensure
  // that the argument is close to 1.
  // Note that RHO is chosen to be an exact floating-point value.
  static const double RHO = 9. / 8;
  static const Affine1Jet LOG_RHO(0.11778303565638345454,
				  0., 0., 0., Affine1Jet::EPS);
  double center_log = log(v._f0) / log(RHO);
  int num_factors = (int) floor(fabs(center_log));
  if( center_log > 0.0 ){
    for( int i=0; i<num_factors; i++ ){
      v /= RHO;
    }
  } else if( center_log < 0.0 ){
    for( int i=0; i<num_factors; i++ ){
      v *= RHO;
    }
  }
  
  // The computation below depends on a Taylor approximation of log
  // around the point x=1, so to compute the error term correctly we
  // need to know if 1 is less than all points in the range of v.
  // I.e. we need to know if v._f0 - sum(|v._dfi|) - v.error <= 1.
  double size = (fabs(v._df1) + fabs(v._df2)) + (fabs(v._df3) + v._error);
  bool mightnt_exceed_one =
    v._f0 <= (1.0 + 3 * Affine1Jet::EPS) * (size + 1.0);

  // We're using the equation log(x) == (x-1)-(x-1)^2/2 +
  // (x-1)^3/(3*c^3) where c is between 1 and x. We choose a
  // worst-case scenario for c. 

  // Compute the "Taylor error".
  double error;
  double cubic = fabs(v._f0 - 1.0) + size;
  if( mightnt_exceed_one ){
    double c = v._f0 - size;
    error = ((v._error + (fabs(v._f0 - 1.0) * v._error)) + 0.5*(size * size)) +
      (cubic * cubic) * (cubic / ((3.0 * c) * (c * c)));
  } else {
    error = ((v._error + (fabs(v._f0 - 1.0) * v._error)) + 0.5*(size * size)) +
      (cubic * cubic) * (cubic / 3.0);
  }

  // Next, add the floating point error.
  error += Affine1Jet::EPS_by_2 * ((((fabs(v._f0 - 1.0) +
			  0.5 * fabs((v._f0 - 1.0) * (v._f0 - 1.0))) +
			 (fabs((v._f0 - 1.0) - 0.5 * ((v._f0 - 1.0) * (v._f0 - 1.0))) +
			  fabs(v._df1 * (2.0 - v._f0)))) +
			((fabs(v._df2 * (2.0 - v._f0)) +
			  fabs(v._df3 * (2.0 - v._f0))) +
			 fabs(v._df1) * fabs(2.0 - v._f0))) +
		       ((fabs(v._df2) * fabs(2.0 - v._f0) +
			 fabs(v._df3) * fabs(2.0 - v._f0)) +
			(1.0 + 2 * Affine1Jet::EPS) *
			(fabs(v._f0 - 1.0) * fabs(v._f0 - 1.0))));

  // Finally, account for the error from computing the errors.
  if( mightnt_exceed_one ){
    error *= 1.0 + 6 * Affine1Jet::EPS;
  } else {
    error *= 1.0 + 5 * Affine1Jet::EPS;
  }

  Affine1Jet result((v._f0 - 1.0) - 0.5 * ((v._f0 - 1.0) * (v._f0 - 1.0)),
		    v._df1 * (2.0 - v._f0),
		    v._df2 * (2.0 - v._f0),
		    v._df3 * (2.0 - v._f0),
		    error);
  
  // Undo the changes we made to v to ensure that v was close to 1:
  if( center_log > 0.0 ){
    result += (LOG_RHO * num_factors);
  } else if( center_log < 0.0 ){
    result -= (LOG_RHO * num_factors);
  }

  return result;
}

Affine1Jet Max0( const Affine1Jet & v )
{
  // We wish to compute an Affine1Jet corresponding to Max(v,0). But
  // if 0 is in the range of v to begin with, then Max(v,0) is very
  // much not a linear function. The best we can do is approximate
  // Max(v,0) with a constant function and use a very large error
  // term, and hope that's good enough.
  if( v.mayContainZero() ){
    double size = v._f0 + ((fabs(v._df1)+fabs(v._df2))+(fabs(v._df3)+v._error));
    size *= 1 + 3 * Affine1Jet::EPS;
    size /= 2; // Exact operation

    return Affine1Jet(size, 0.0, 0.0, 0.0, size);
  }

  // If 0 isn't in the range of v, things are much simpler.

  if( v._f0 >= 0.0 ){
    // Note that v._f0 == 0 should actually be covered by the previous
    // case.
    return v;
  }

  // At this point, the range of v is strictly below 0.
  return Affine1Jet(0.0, 0.0, 0.0, 0.0, 0.0);
}

Affine1Jet ArcTan( Affine1Jet v )
{
  // Like the logarithm, all of the derivatives of ArcTan are rational
  // polynomials. So we will use a Taylor approximation to compute the
  // arctangent and use Taylor's theorem to bound the difference
  // between the arctangent and the approximation.

  // ArcTan is odd, might as well use the symmetry.
  bool negated = false;
  if( v._f0 < 0 ){
    negated = true;
    v = -v;
  }

  // Since we're only using a crude polynomial approximation to arctan
  // we prefer to deal with arguments as close to zero as
  // possible. We start with the formula
  //   ArcTan(a) + ArcTan(b) = ArcTan((a+b)/(1-a*b))
  // Let c = (a+b)/(1-a*b) and solve for b:
  //   ArcTan(c) = ArcTan(a) + ArcTan((c-a)/(a*c+1))
  // If c is positive and a is a positive constant less than c then
  // the second argument on the right will be smaller than c. Repeat
  // until you end up with an argument smaller than a.
  // Currently as a constant I use a number with an exact binary
  // representation which is smaller than tan(Pi/12).
  static const double RHO = 137. / 512;
  static const Affine1Jet ARCTAN_RHO(0.26145314506663815251,
				     0., 0., 0., Affine1Jet::EPS);
  int num_translations = 0;
  while( v._f0 > RHO ){
    num_translations++;
    v = (v-RHO) / (v*RHO+1);
  }

  // We use the equation arctan(x) == x-x^3/3 +
  // x^4*(c-2*c^3)/(2*(c^2+1)^4) where c is between 0 and x, and
  // choose a worst-case scenario for c.

  // To find the worst-case scenario for c, we need to know whether or
  // the range of |v| extends to the right of the global maximum of
  // the function (c-2*c^3)/(2*(c^2+1)^4). This isn't pretty, but it
  // works.
  double size = (fabs(v._df1) + fabs(v._df2)) + (fabs(v._df3) + v._error);
  bool includes_max =
    300466. / 1048576 <= (1.0 + 3 * Affine1Jet::EPS) * (fabs(v._f0) + size);

  // Now we can compute the "Taylor error".
  double error;
  error = (v._error + v._error * fabs(v._f0 * v._f0)) +
    (fabs(v._f0) * (size * size) + (size * size) * (size / 3.0));
  if( includes_max ){
    double c = 91575. / 1048576;
    double c2p1 = c * c + 1.0;
    error += ((((fabs(v._f0) + size) * (fabs(v._f0) + size)) *
	       ((fabs(v._f0) + size) * (fabs(v._f0) + size))) *
	      ((c - (2.0 * c) * (c * c)) / 2.0)) /
      ((c2p1 * c2p1) * (c2p1 * c2p1));
  } else {
    double c = fabs(v._f0) + size;
    double c2p1 = c * c + 1.0;
    error += (((c * c) * (c * c)) * ((c - (2.0 * c) * (c * c)) / 2.0)) /
      ((c2p1 * c2p1) * (c2p1 * c2p1));
  }

  // Next, add the floating point error.
  error += Affine1Jet::EPS_by_2 * (((2 * (1 + 2 * Affine1Jet::EPS) *
				     fabs(v._f0 * v._f0) * fabs(v._f0 / 3.0) +
				     fabs((v._f0 * v._f0) * (v._f0 / 3.0))) +
				    (fabs(v._f0 - (v._f0 * v._f0) * (v._f0 / 3.0)) +
				     fabs(v._df1) * fabs(v._f0 * v._f0))) +
				   (((fabs(v._df1 * (v._f0 * v._f0)) +
				      fabs(v._df1 - v._df1 * (v._f0 * v._f0))) +
				     (fabs(v._df2) * fabs(v._f0 * v._f0) +
				      fabs(v._df2 * (v._f0 * v._f0)))) +
				    ((fabs(v._df2 - v._df2 * (v._f0 * v._f0)) +
				      fabs(v._df3) * fabs(v._f0 * v._f0)) +
				     (fabs(v._df3 * (v._f0 * v._f0)) +
				      fabs(v._df3 - v._df3 * (v._f0 * v._f0))))));

  // Finally, account for the error from computing the errors.
  if( includes_max ){
    error *= 1.0 + 6 * Affine1Jet::EPS;
  } else {
    error *= 1.0 + 7 * Affine1Jet::EPS;
  }

  Affine1Jet result(v._f0 - (v._f0 * v._f0) * (v._f0 / 3.0),
		    v._df1 - v._df1 * (v._f0*v._f0),
		    v._df2 - v._df2 * (v._f0*v._f0),
		    v._df3 - v._df3 * (v._f0*v._f0),
		    error);

  // Undo the transformations that we performed to get the argument
  // close to zero.
  while( num_translations > 0 ){
    result += ARCTAN_RHO;
    num_translations--;
  }

  if( negated ){
    result = -result;
  }

  return result;
}
